#include "fmt_opts.h"
#include <math.h>
#include <assert.h>

strmap formatting_options::styles;

std::string formatting_options::get_par_str_closing_token() const
{
   if (listActive == true)
      return "</li>\n";
   else
      return "</p>\n";
}


std::string formatting_options::get_par_str() const
{
   std::string style;
   std::string styleAddedToBlock;
   switch (papAlign)
   {
   case formatting_options::align_right:
      style+="text-align:right;";
      // Qt doesn't support the "text-align" css property. Thus add the alignment as a block attribute as well.
      styleAddedToBlock +=" align=\"right\"";
      break;
   case formatting_options::align_center:
      style+="text-align:center;";
      // Qt doesn't support the "text-align" css property. Thus add the alignment as a block attribute as well.
      styleAddedToBlock +=" align=\"center\"";
      break;
   case formatting_options::align_justify:
      style+="text-align:justify;";
      // Qt doesn't support the "text-align" css property. Thus add the alignment as a block attribute as well.
      styleAddedToBlock +=" align=\"justify\"";
      break;
   case formatting_options::align_left:
   case formatting_options::align_error:
      // Nothing to do
      break;
   }
   if (papFirst!=0)
   {
      style+="text-indent:";
      style+=from_int(papFirst);
      style+="pt;";
   }
   if (papLeft!=0)
   {
      style+="margin-left:";
      style+=from_int(papLeft);
      style+="pt;";

      // Qt uses an own block attribute for text indentation. We need to use this one to be compatible.
      style += "-qt-block-indent:" + from_int(rint(papLeft / 36)) + ";";
      //styleAddedToBlock += "margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:" + from_int(rint(papLeft / 36)) + "; text-indent:0px;";
   }
   if (papRight!=0)
   {
      style+="margin-right:";
      style+=from_int(papRight);
      style+="pt;";
   }
   if (papBefore!=0)
   {
      style+="margin-top:";
      style+=from_int(papBefore);
      style+="pt;";
   }
   if (papAfter!=0)
   {
      style+="margin-bottom:";
      style+=from_int(papAfter);
      style+="pt;";
   }
   if (listActive == true)
   {
      style += "-qt-list-indent: 1;";
   }
   if (style.empty() && styleAddedToBlock.empty())
      return std::string("<p>");
   else
   {
      //return std::string("<p class=\"") + formatting_options::get_style_id(style) + "\">";
      std::string result;
      result += listActive ? "<li " : "<p ";
      result += styleAddedToBlock + std::string(" class=\"") + formatting_options::get_style_id(style) + "\">";
      return result;
   }
}

std::string formatting_options::get_style_id(const std::string &style)
{
   strmap::iterator i_style = styles.find(style);
   if (i_style == styles.end())
   {
      i_style = styles.insert(strmap::value_type(style, std::string("cls") + from_int(styles.size()))).first;
   }
   return i_style->second;
}

std::string formatting_options::get_styles()
{
   std::string result;
   for (strmap::const_iterator i = styles.begin(); i != styles.end(); ++i)
   {
      result += std::string(".") + i->second + " {" + i->first + "}\n";
   }
   return result;
}

std::string formatter::format(const formatting_options &_opt)
{
   formatting_options last_opt, opt(_opt);
   std::string result;
   if (!opt_stack.empty())
   {
      int cnt=0;
      fo_deque::reverse_iterator i;
      for (i=opt_stack.rbegin(); i!=opt_stack.rend(); ++i)
      {         
         if (*i==opt)
            break;
         ++cnt;
      }
      if (cnt==0)
         return "";
      if (i!=opt_stack.rend())
      {
         while (cnt--)
         {
            result+="</span>";
            opt_stack.pop_back();
         }
         return result;
      }
      last_opt=opt_stack.back();
   }
   if (last_opt.chpVAlign!=formatting_options::va_normal 
       && last_opt.chpVAlign!=opt.chpVAlign)
   {
      int cnt=0;
      fo_deque::reverse_iterator i;
      for (i=opt_stack.rbegin(); i!=opt_stack.rend(); ++i)
      {         
         if (i->chpVAlign==formatting_options::va_normal)
            break;
         ++cnt;
      }
      while (cnt--)
      {
         result+="</span>";
         opt_stack.pop_back();
      }
      last_opt=opt_stack.empty()?formatting_options():opt_stack.back();
   }
   std::string style;
   if (opt.chpBold!=last_opt.chpBold)
   {
      style+="font-weight:";
      style+=opt.chpBold?"bold":"normal";
      style+=";";
   }
   if (opt.chpAllCaps!=last_opt.chpAllCaps)
   {
      style+="text-transform:";
      style+=opt.chpAllCaps?"uppercase":"none";
      style+=";";
   }
   if (opt.chpItalic!=last_opt.chpItalic)
   {
      style+="font-style:";
      style+=opt.chpItalic?"italic":"normal";
      style+=";";
   }
   if (opt.chpUnderline!=last_opt.chpUnderline)
   {
      style+="text-decoration:";
      style+=opt.chpUnderline?"underline":"none";
      style+=";";
   }
   if (opt.chpVAlign!=formatting_options::va_normal)
      opt.chpFontSize=(int)(0.7*(opt.chpFontSize?opt.chpFontSize:24));
   if (opt.chpFontSize!=last_opt.chpFontSize)
   {
      style+="font-size:";
      int iFontSizePt = opt.chpFontSize / 2;

#ifdef WIN32
      // On Windows we need to reduce the font size by one to get closer to the original text layout in the TE edit based version.
      style+=from_int(iFontSizePt - 1);
      style+="pt;";
#else
      // On Mac the text layout looks looks quite good with the original TE edit font size
      style+=from_int(iFontSizePt);
      style+="pt;";
#endif
   }
   if (opt.chpVAlign!=last_opt.chpVAlign)
   {
      style+="vertical-align:";
      style+=opt.chpVAlign==formatting_options::va_sub?"sub":"super";
      style+=";";
   }
   if (opt.chpFColor!=last_opt.chpFColor)
   {
      style+="color:";
      style+=opt.chpFColor.r>=0?"#"+hex(opt.chpFColor.r&0xFF)
                                  +hex(opt.chpFColor.g&0xFF)
                                  +hex(opt.chpFColor.b&0xFF)
                              :"WindowText";
      style+=";";
   }
   if (opt.chpBColor!=last_opt.chpBColor)
   {
      style+="background-color:";
      style+=opt.chpBColor.r>=0?"#"+hex(opt.chpBColor.r&0xFF)
                                  +hex(opt.chpBColor.g&0xFF)
                                  +hex(opt.chpBColor.b&0xFF)
                              :"Window";
      style+=";";
   }
   if (opt.chpHighlight!=last_opt.chpHighlight)
   {
      style+="background-color:";
      switch (opt.chpHighlight)
      {
      case 0: style+="Window"; break;
      case 1: style+="black"; break;
      case 2: style+="blue"; break;
      case 3: style+="aqua"; break;
      case 4: style+="lime"; break;
      case 5: style+="fuchsia"; break;
      case 6: style+="red"; break;
      case 7: style+="yellow"; break;
      case 9: style+="navy"; break;
      case 10: style+="teal"; break;
      case 11: style+="green"; break;
      case 12: style+="purple"; break;
      case 13: style+="maroon"; break;
      case 14: style+="olive"; break;
      case 15: style+="gray"; break;
      case 16: style+="silver"; break;
      }
      style+=";";
   }
   if (opt.chpVShift!=last_opt.chpVShift)
   {
      style += "position:relative;top:" + from_int(opt.chpVShift / 2) + "pt;";
   }
   if (opt.chpFont!=last_opt.chpFont)
   {
      style+="font-family:'";
      style+=opt.chpFont.name.empty()?"serif":opt.chpFont.name;
      style+="'";
/*
      // [FABIAN]: Disable the family suffix since the font names would then not match the names in the font manager anymore
      switch (opt.chpFont.family)
      {
      case font::ff_serif: style+=", serif"; break;
      case font::ff_sans_serif: style+=", sans-serif"; break;
      case font::ff_cursive: style+=", cursive"; break;
      case font::ff_fantasy: style+=", fantasy"; break;
      case font::ff_monospace: style+=", monospace"; break;
      case font::ff_none: break;
      }
*/
      style+=";";
   }
   opt_stack.push_back(opt);
   return result + "<span class=\"" + formatting_options::get_style_id(style) + "\">";
}

std::string formatter::close()
{
   std::string result;
   for (fo_deque::iterator i=opt_stack.begin(); i!=opt_stack.end(); ++i)
      result+="</span>";
   return result;
}
